#ifndef __INTERNALS_H__
#define __INTERNALS_H__

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/iom128.h>
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "serial.h"
#include "semphr.h"
#include "sys.h"
#include "opcodes.h"


#define UTIME(x) ((x) << 0x1)

/* Internal constants */

/* Motor ct */
enum motorCfgCt {		
	L_WHEEL = 0,
	R_WHEEL = 1,
	MOTOR_ADDR_CFG = 0x0000,	// if it change, change EEPROM data after
};

/* Lateral IR ct */
enum irLatCfgCt { 
	IRL_CH = ADCH_2,
	IRR_CH = ADCH_0,

	ADCx32 = 0x05,
	ADCx64 = 0x06,

	TSTABLE = 0x0F,
};

/* Frontal IR ct */
enum irFrontCfgCt {
	PERIOD = UTIME(26), 
	HPERIOD = PERIOD/2, 
	
	IRF_LEFT = 1, 
	IRF_RIGHT = 0,

	SH_FACTOR = 2,
	OFF_FACTOR = 100,

	/* States of pulse seq. */
	PUL_ST1 	= 0x0,  
	DELAY_ST1 	= 0x1,
	XPH			= 0x2,
	PUL_ST2 	= 0x3,   
	DELAY_ST2 	= 0x4,
	END_ST		= 0x5,
};


enum intCfgCt {
	L_LEVEL = 0x0,
	CH_LEVEL = 0x1, 
	F_EDGE = 0x2,
	R_EDGE = 0x3,
};


enum ldrCfgCt {
	LDRL_CH = ADCH_3,
	LDRR_CH = ADCH_1,

	LDR_WAIT = 2,
};


enum ledCfgCt {
	LED_MSK = 0x3F,
	
	LOW_POS = 0,
	LOWER_LEDS_MSK = (0x7 << LOW_POS),

	UP_POS = 3,
	UPPER_LEDS_MSK = (0x7 << UP_POS),

	SET_LEDS = 0,
	CLEAR_LEDS = 1,

	TOG_LEDS = 0,
	TIME_LEDS = 1,
};


enum spkCfgCt {
	SPK_MSK = 0x7,

	TIME_SPK = 0,
	CFG_SPK = 1,
};


enum lfwCfgCt {
	FEL = (0x1 << PINC5),	
	FRL = (0x1 << PINC7),
	BITL = (0x1 << 0),

	FECL = (0x1 << PINC1),
	FRCL = (0x1 << PINC3),
	BITCL = (0x1 << 1),

	FECR = (0x1 << PINC2),
	FRCR = (0x1 << PINC0),
	BITCR = (0x1 << 2),

	FER	= (0x1 << PINC6),
	FRR = (0x1 << PINC4), 
	BITR = (0x1 << 3),

	
	FLEN = 0x4,
	EMITTER = 0x0,
	RECEIVER = 0x1,
	BITPOS = 0x2,
	
	LFW_WAIT = 2, 
};


enum micCfgCt {
	MIC_CH = ADCH_3,
};


/* Misc ct */
enum  miscCfgCt {
	BSIZE = 8, 
	
	NO_WAIT = 0,
	UNDF_WAIT = portMAX_DELAY,    
};



/* Internal datatypes */
typedef struct coreData coreCtrl_t;
typedef struct recvPkg recvPkg_t;
typedef struct sendPkg sendPkg_t;
typedef struct ledCtrl ledCtrl_t;
typedef struct sensorCtrl sensorCtrl_t;
typedef struct motorCtrl motorCtrl_t;
typedef struct sharedCtrl sharedCtrl_t;
typedef portBASE_TYPE (*vecParse_t)(coreCtrl_t *, uint8_t );
typedef portBASE_TYPE (*vecCtrl_t)(sensorCtrl_t *, sendPkg_t *);
typedef portBASE_TYPE (*vecCmplCtrl_t)(motorCtrl_t *, recvPkg_t *, sendPkg_t *);


struct ledCtrl {
	uint8_t oscLed;
	uint8_t pre; 
};

struct sensorCtrl {
	xQueueHandle sensorQueue;
	xSemaphoreHandle syncTask;
	sharedCtrl_t *shCtrl;
};

struct motorCtrl {
	volatile ledCtrl_t led;
	int16_t pn[2];

	xQueueHandle motorQueue;
	xSemaphoreHandle syncTask;
	sharedCtrl_t *shCtrl;
};

struct sharedCtrl {
	xSemaphoreHandle syncSendSerial;
};

/* Internal structs */
struct coreData {
	motorCtrl_t mCtrl;
	sensorCtrl_t sCtrl;
	sharedCtrl_t shCtrl;
};

struct recvPkg {
	uint8_t op;

	union {
		int8_t wheelOff[2];
		uint8_t argList[2];	
	};

	union {
		uint8_t arg;
		uint8_t delay;
	};
};


struct sendPkg {
	uint8_t code;

	union {
		uint8_t len;
		uint8_t err;
	};

	uint8_t *data;
};


/* Functions and Proc prototypes */
void initSystem(void);
void serialTask(void *pvParameters);
void motorTask(void *pvParameters);
void sensorTask(void *pvParameters);


portBASE_TYPE sendSerialString(xSemaphoreHandle sync, const void *buff, const uint16_t len);
portBASE_TYPE sendSerialPkg(xSemaphoreHandle sync, const sendPkg_t const *sndPkg);
portBASE_TYPE hex2str(const uint8_t * const num, const uint8_t len);


#ifdef __DEBUG__
	extern void vApplicationStackOverflowHook( xTaskHandle *pxTask, signed portCHAR *pcTaskName );
#endif



#endif
