#include "RUABParams.h"


static inline void setupSensorsCfg(void);
static inline void setupMotorsCfg(motorCtrl_t *);


void initSystem(void)
{
	coreCtrl_t *cCtrl;
	xSemaphoreHandle s;

	xSerialPortInitMinimal( SERIAL_BAUDRATE, SERIAL_RS232_QUEUE_SIZE );

	cCtrl = (coreCtrl_t *) pvPortMalloc( sizeof(coreCtrl_t) );
	
	cCtrl->mCtrl.motorQueue = xQueueCreate(MOTOR_QUEUE_LENGTH, MOTOR_QUEUE_ITEM_SIZE);
	cCtrl->sCtrl.sensorQueue = xQueueCreate(SENSOR_QUEUE_LENGTH, SENSOR_QUEUE_ITEM_SIZE);
	
	vSemaphoreCreateBinaryLock( cCtrl->mCtrl.syncTask );
	vSemaphoreCreateBinaryLock( cCtrl->sCtrl.syncTask );
	s = cCtrl->sCtrl.syncTask;

	cCtrl->shCtrl.syncSendSerial = xSemaphoreCreateMutex();
	cCtrl->mCtrl.shCtrl = &(cCtrl->shCtrl);
	cCtrl->sCtrl.shCtrl = &(cCtrl->shCtrl);

	setupMotorsCfg( &(cCtrl->mCtrl) );
	setupSensorsCfg();

	xTaskCreate(serialTask, (signed char *) "serialTask", configMINIMAL_STACK_SIZE, ((void *) cCtrl), SERIAL_TASK_PRIORITY, NULL);
	xTaskCreate(motorTask, (signed char *) "motorTask", configMINIMAL_STACK_SIZE, ((void *) &(cCtrl->mCtrl)), MOTOR_TASK_PRIORITY, NULL);
	xTaskCreate(sensorTask, (signed char *) "sensorTask", configMINIMAL_STACK_SIZE, ((void *) &(cCtrl->sCtrl)), SENSOR_TASK_PRIORITY, NULL);
}


/*
	@setupMotorsCfg(motorCtrl_t *mCtrl): Initialize timer1 config. to set ready PWM signal
		
	Generate PWM-Waveform:
		- Select Mode: PWM Phase and Frequency Mode (50 Hz / 20 ms PWM period)
		- Prescaler: 8  
		
		- Register A as TOP. Clear timer counter on match
		- Registers B and C reset signal-PWM on match (down to 0)
*/
static inline void setupMotorsCfg(motorCtrl_t *mCtrl)
{
	mCtrl->pn[L_WHEEL] = eeprom_read_word((uint16_t *) MOTOR_ADDR_CFG );
	mCtrl->pn[R_WHEEL] = eeprom_read_word((uint16_t *) (MOTOR_ADDR_CFG + 2));

	ICR1 = MOTOR_PWM_PERIOD;
	OCR1B = mCtrl->pn[L_WHEEL];
	OCR1A = mCtrl->pn[R_WHEEL];	

	DDRB |= ( (0x1 << PINB5) | (0x1 << PINB6) );
	
	TCCR1A |= ( (0x1 << COM1A1) | (0x1 << COM1B1) );
	TCCR1B |= ( (0x1 << WGM13) | (0x2 << CS10) );
}


/*
	@setupSensorsCfg(void): Initialize difenrent kinds of config. related with sensors
*/
static inline void setupSensorsCfg(void)
{
	/* 	
		Port E: Control LDRs, control IRF, output IRF, control IRL
		 
		Port E Pin 2, control inputs LDRs (POUT)
		Port E Pin 4,5 control frontal IR inputs (OC3B & OC3C)
		Port E Pin 6 output frontal IR (EINT6)  
		Port E Pin 7, control inputs lateral IR (POUT)
	*/
	DDRE |= ( 0x1 << PINE2 | 0x1 << PINE4 | 0x1 << PINE5 | 0x1 << PINE7 );
	DDRE &= ~( 0x1 << PINE6 );
	PORTE = PINE & ~( 0x1 << PINE2 | 0x1 << PINE4 | 0x1 << PINE5 | 0x1 << PINE7 );

	/*
		Port B: Leds & Speaker

		Port B Pin 0,1,2 control leds (POUT)
		Port B Pin 7 control speaker (OC2)
	*/
	DDRB |= (0x1 << PINB0 | 0x1 << PINB1 | 0x1 << PINB2 | 0x1 << PINB7);
	PORTB = PINB & ~( 0x1 << PINB0 | 0x1 << PINB1 | 0x1 << PINB2 | 0x1 << PINB7 ); 
	
	TCCR2 = (0x1 << WGM21 | 0x1 << COM20 | 0x0);
	OCR2 = 0x0;

	/*
		Port G: Leds

		Port G Pin 0,1,2 control leds (POUT)
	*/
	DDRG |= (0x1 << PING0 | 0x1 << PING1 | 0x1 << PING2);
	PORTG = PING & ~(0x1 << PING0 | 0x1 << PING1 | 0x1 << PING2);

	/*
		Port C: Control Line Follower

		Port C Pin 1, 2, 5, 6: Emitter (Center-Right, Center-Left, Right, Left) (POUT)
		Port C Pin 0, 3, 4, 8: Receptor (Center-Left, Center-Right, Left, Right) (PIN)
	*/
	DDRC = (0x1 << PINC1 | 0x1 << PINC2 | 0x1 << PINC5 | 0x1 << PINC6 );
	PORTC = PINC & ~(0x1 << PINC1 | 0x1 << PINC2 | 0x1 << PINC5 | 0x1 << PINC6);
}


#ifdef __DEBUG__

void vApplicationStackOverflowHook( xTaskHandle *pxTask, signed portCHAR *pcTaskName )
{
	int i;

	for(i = 0; i< 500; i++)
	{
		if((int)pxTask & 0x4)
			break;
		else
			pxTask = (xTaskHandle *) pcTaskName;
	}
}

#endif



