#include "RUABParams.h"


static const char checkTestMsg[] = "S-Task: OK\n";


/* Shared data */
static volatile uint8_t pul, st = PUL_ST1;
static volatile uint8_t *mem;
static volatile xSemaphoreHandle sph;


/* Functions & Procedures Prototypes */
static uint8_t sampleADC(adch_t ch);
static void stPulse(void);
static void runIRF(uint8_t *data);


/* Inline Functions & Procedures Prototypes */
static inline void setupSensorsParams(sensorCtrl_t *);
static inline void initIRLateral(void);
static inline void stopIRLateral(void);
static inline void initIRFrontal(void);


/* Sensor - Command Routines Prototypes */
static portBASE_TYPE testTaskSOp(sensorCtrl_t *, sendPkg_t *);
static portBASE_TYPE writeCfgMotorOp(sensorCtrl_t *, sendPkg_t *);
static portBASE_TYPE irLateralOp(sensorCtrl_t *, sendPkg_t *);
static portBASE_TYPE irFrontalOp(sensorCtrl_t *, sendPkg_t *);
static portBASE_TYPE ldrOp(sensorCtrl_t *, sendPkg_t *);
static portBASE_TYPE lfwOp(sensorCtrl_t *, sendPkg_t *);
static portBASE_TYPE micOp(sensorCtrl_t *, sendPkg_t *);
static portBASE_TYPE allOp(sensorCtrl_t *, sendPkg_t *);
static portBASE_TYPE nopOp(sensorCtrl_t *, sendPkg_t *);


/* Sensor - Vector Command Routines */
static const vecCtrl_t VCR[] = { 
	testTaskSOp,			// Opcode TSTS
	writeCfgMotorOp, 		// Opcode WCM
	irLateralOp, 			// Opcode IRL
	irFrontalOp,			// Opcode IRF
	ldrOp, 					// Opcode LDR
	lfwOp,					// Opcode LFW
	micOp,					// Opcode MIC
	allOp,		 			// Opcode ALL
	nopOp, 					// 
	nopOp, 					// 
	nopOp, 					// 
	nopOp, 					// 
	nopOp, 					// 
	nopOp, 					// 
	nopOp, 					// 
	nopOp, 					// 
	nopOp, 					// 
	nopOp, 					// 
	nopOp, 					// 
};



/*----------------------------------------------------------------------------------*/
/*																					*/
/*								Procedures / Functions 						 		*/
/*																					*/
/*----------------------------------------------------------------------------------*/

void sensorTask(void *pvParameters)
{
	sensorCtrl_t *sCtrl = ((sensorCtrl_t *) pvParameters);
	sendPkg_t sndPkg;
	uint8_t buff[BSIZE];

	setupSensorsParams(sCtrl);
	xSemaphoreGive(sCtrl->syncTask);
	for(; ; sndPkg.data = buff)
	{
		if( xQueueReceive(sCtrl->sensorQueue, &(sndPkg.code), SENSOR_QUEUE_WAIT) == pdTRUE )
		{
			if( (sndPkg.code > OP_THRESHOLD) || (VCR[sndPkg.code - SENSOR_OPCODE_OFFSET](sCtrl, &sndPkg) != pdTRUE) )
			{
				sndPkg.err = sndPkg.code; 
				sndPkg.code = 0x0;
			}

			sendSerialPkg((sCtrl->shCtrl)->syncSendSerial, &sndPkg);
		}
	}
}


static uint8_t sampleADC(adch_t ch)
{
	/* 
		Select channel & Config ADC
			- AVcc Voltage reference, 
			- ADLAR: Left Adjust Result, 
			- Channel: ch
	*/
	ADMUX = ( 0x1 << REFS0 | 0x1 << ADLAR | (ch & 0x7) );

	/* Preescaler XTAL/64 (8-bit sample) */
	ADCSRA = ( 0x1 << ADEN | 0x1 << ADSC | ADCx64 );

	for( ; !(ADCSRA & (0x1 << ADIF)); ) ;
	ADCSRA |= (0x1 << ADIF);

	return ADCH;
}


static void stPulse(void)
{
	register uint8_t side = (st >> 1);
	signed portBASE_TYPE swc;
	 
	switch(st)
	{
		case XPH: st++;
		case PUL_ST2:
		case PUL_ST1:
			if( pul < (NPULSES - 1) ) pul++;
			else 
			{
				pul = 0;
				OCR3A = DELAY;
				(&OCR3C)[1-side] = 0;

				st++;	// Next state
			}
			break;


		case DELAY_ST1:
			/* Set up Timer Counter, Periode (OCR3A) and reconfig. Timer */
			OCR3A = PERIOD;
			OCR3C = HPERIOD;
			TCCR3A = ((0x1 << WGM31) | (0x1 << WGM30) | (0x1 << COM3C1) );

			st++;
			break;


		case DELAY_ST2:
			TCCR3A = ((0x1 << WGM31) | (0x1 << WGM30));
			st++;
			break;


		case END_ST:
		default:
			/* Stop Timer */
			TCCR3B = 0x0;
			ETIMSK &= ~( 0x1 << OCIE3A );
			ETIFR |= ( 0x1 << OCIE3A | 0x1 << OCIE3B  | 0x1 << OCIE3C | 0x1 << ICF3 | 0x1 << TOV3 );

			xSemaphoreGiveFromISR(sph, &swc);
			if(swc == pdTRUE)
				taskYIELD();

			st = PUL_ST1;
			break;
	}	
}


static void timeHigh(void)
{
	mem[(st > XPH) ? 1 : 0] = \
		(uint8_t) ((( ((pul) ? pul : NPULSES) * PERIOD) + TCNT3) >> SH_FACTOR);	
}


static void runIRF(uint8_t *data)
{
	mem = data;
	mem[0] = ACTTIME;
	mem[1] = ACTTIME;
		
	initIRFrontal(); 	
}



/*----------------------------------------------------------------------------------*/
/*																					*/
/*					Inline Procedures: Config/Init/Stop instructions			 	*/
/*																					*/
/*----------------------------------------------------------------------------------*/

static inline void setupSensorsParams(sensorCtrl_t *sCtrl)
{
	/* Sync. Semaphore*/
	sph = sCtrl->syncTask;

	/* Functions for interrupts */
	timer3_compA_hook_isr = stPulse;
	int6_hook_isr = timeHigh;
}


static inline void initIRLateral(void)
{
	uint8_t tmp;

	PORTE = PINE | (0x1 << PINE7);

	for(tmp = 0; tmp != TSTABLE; tmp++) ;
}


static inline void stopIRLateral(void)
{
	PORTE = PINE & ~(0x1 << PINE7);
	ADCSRA &= ~(0x1 << ADEN); 
}


static inline void initIRFrontal(void)
{
	/* Config. timer and interrupt and start timer */
	OCR3A = PERIOD;
	OCR3B = HPERIOD;

	ETIMSK |= (0x1 << OCIE3A) ; 
	
	// Fast PWM (OCR3A as TOP), Prescaler 8
	TCCR3A = ( (0x1 << WGM31) | (0x1 << WGM30) | (0x1 << COM3B1) );
	TCCR3B = (0x1 << WGM33) | (0x1 << WGM32) | 0x2;	
}



/*----------------------------------------------------------------------------------*/
/*																					*/							 
/*								Available Command Routines 							*/
/*																					*/	
/*----------------------------------------------------------------------------------*/
static portBASE_TYPE testTaskSOp(sensorCtrl_t *sCtrl, sendPkg_t *sndPkg)
{
	(void) sCtrl;

	sndPkg->data = (uint8_t *) checkTestMsg;
	sndPkg->len = sizeof(checkTestMsg);

	return pdTRUE;
}


static portBASE_TYPE writeCfgMotorOp(sensorCtrl_t *sCtrl, sendPkg_t *sndPkg)
{
	(void) sCtrl;

	sndPkg->len = WCM_LEN_DATA;
	
	portENTER_CRITICAL();

	eeprom_write_word((uint16_t *) MOTOR_ADDR_CFG, OCR1B);	
	eeprom_write_word((uint16_t *) (MOTOR_ADDR_CFG + 2), OCR1A);

	portEXIT_CRITICAL();

	
	return pdTRUE;
}


static portBASE_TYPE irLateralOp(sensorCtrl_t *sCtrl, sendPkg_t *sndPkg)
{
	(void) sCtrl;

	initIRLateral();
	sndPkg->len = IRL_LEN_DATA;

	/* Sampling */
	(sndPkg->data)[0] = sampleADC(IRL_CH);

	/* This reduce error conversion */
	PORTE = PINE & ~(0x1 << PINE7);
	ADCSRA &= ~(0x1 << ADEN);
	PORTE = PINE | (0x1 << PINE7);

	(sndPkg->data)[1] = sampleADC(IRR_CH);

	stopIRLateral();
	
	return pdTRUE;	
}


static portBASE_TYPE irFrontalOp(sensorCtrl_t *sCtrl, sendPkg_t *sndPkg)
{
	portBASE_TYPE r;

	enableINT(IRQ6, F_EDGE);
	runIRF(sndPkg->data);
	sndPkg->len = IRF_LEN_DATA;
	r = xSemaphoreTake(sCtrl->syncTask, IRF_MAX_WAIT_TIME);
	disableINT(IRQ6);

	return r;


}


static portBASE_TYPE ldrOp(sensorCtrl_t *sCtrl, sendPkg_t *sndPkg)
{
	(void) sCtrl;


	PORTE = PINE | (0x1 << PINE2);
	sndPkg->len = LDR_LEN_DATA;
	
	vTaskDelay(LDR_WAIT);
	(sndPkg->data)[0] = sampleADC(LDRL_CH);
	(sndPkg->data)[1] = sampleADC(LDRR_CH);

	PORTE = PINE & ~(0x1 << PINE2);

	return pdTRUE;
}


static portBASE_TYPE lfwOp(sensorCtrl_t *sCtrl, sendPkg_t *sndPkg)
{
	(void) sCtrl;

	static const uint8_t fer[FLEN][3] = { 	
		{FECL, FRCL, BITCL}, {FER, FRR, BITR},
		{FEL, FRL, BITL}, {FECR, FRCR, BITCR}, 
	};
	uint8_t i;


	sndPkg->data[0] = 0x0F;
	sndPkg->len = LFW_LEN_DATA;

	
	for(i = 0; i < FLEN; i++)
	{
		PORTC = fer[i][EMITTER];
		vTaskDelay(LFW_WAIT);
		sndPkg->data[0] ^= ( (PINC & fer[i][RECEIVER]) ? fer[i][BITPOS] : 0x0);
	}
	PORTC = 0x0;

	/*
	PORTC = (fer[0][EMITTER] | fer[3][EMITTER]);
	vTaskDelay(LFW_WAIT);
	sndPkg->data[0] ^= ( (PINC & fer[0][RECEIVER]) ? 0x1<<0 : 0x0 );
	sndPkg->data[3] ^= ( (PINC & fer[3][RECEIVER]) ? 0x1<<3 : 0x0 );

	PORTC = fer[1][EMITTER];
	vTaskDelay(LFW_WAIT);
	sndPkg->data[1] ^= ( (PINC & fer[1][RECEIVER]) ? 0x1<<1 : 0x0 );

	PORTC = fer[2][EMITTER];
	vTaskDelay(LFW_WAIT);
	sndPkg->data[2] ^= ( (PINC & fer[2][RECEIVER]) ? 0x1<<2 : 0x0 );
	*/
	

	return pdTRUE;
}

static portBASE_TYPE micOp(sensorCtrl_t *sCtrl, sendPkg_t *sndPkg)
{
	(void) sCtrl;

	sndPkg->len = MIC_LEN_DATA;
	(sndPkg->data)[0] = sampleADC(MIC_CH);
	
	return pdTRUE;
}


/*
	data[0 - 1]: IR Lateral
	data[2 - 3]: Frontal IR samples
	data[4 - 5]: LDR samples
*/
static portBASE_TYPE allOp(sensorCtrl_t *sCtrl, sendPkg_t *sndPkg)
{	
	(void) sndPkg;
	/*runIRF( &(ctrl->buff[2]) );
	ldrOP( &(ctrl->buff[4]) );
	irLateralOp( ctrl->buff );*/

	return xSemaphoreTake(sCtrl->syncTask, IRF_MAX_WAIT_TIME);
}


static portBASE_TYPE nopOp(sensorCtrl_t *sCtrl, sendPkg_t *sndPkg)
{
	(void) sCtrl;
	sndPkg->len = NOP_LEN_DATA;

	return pdTRUE;
}



